#!/usr/bin/env python
# filename: bar.py


#
# Copyright (c) 2022 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import math
from collections import Counter
from typing import Iterable, Optional, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from natsort import natsorted

from ..core.sequence import Sequence
from ..utils.color import get_cmap
from .data import process_input_data

__all__ = ["bar"]


def bar(
    x: Union[str, Iterable, None] = None,
    y: Union[str, Iterable, None] = None,
    hue: Union[str, Iterable, None] = None,
    data: Optional[pd.DataFrame] = None,
    sequences: Optional[Iterable[Sequence]] = None,
    order: Optional[Iterable] = None,
    hue_order: Optional[Iterable] = None,
    palette: Union[dict, Iterable, None] = None,
    color: Union[str, Iterable, None] = None,
    alt_color: Union[str, Iterable] = "#D3D3D3",
    normalize: bool = False,
    highlight: Union[str, Iterable, None] = None,
    highlight_color: Union[str, Iterable, None] = None,
    orientation: str = "vertical",
    plot_kwargs: Optional[dict] = None,
    legend_kwargs: Optional[dict] = None,
    hide_legend: bool = False,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    xlabel_fontsize: Union[int, float] = 16,
    ylabel_fontsize: Union[int, float] = 16,
    xtick_labelsize: Union[int, float] = 14,
    ytick_labelsize: Union[int, float] = 14,
    xtick_labelrotation: Union[int, float] = 0,
    ytick_labelrotation: Union[int, float] = 0,
    ax: Optional[mpl.axes.Axes] = None,
    show: bool = False,
    figsize: Optional[Iterable] = None,
    figfile: Optional[str] = None,
):
    """
    Produces a bar plot of categorical data. For data with distinct batches, a stacked
    bar plot will be constructed.

    Parameters
    ----------
    x : str or list, optional
        Name of a column in `data` or an iterable of values to be plotted on the
        x-axis. Required if `orientation` is ``"vertical"``. If only `x` is provided,
        the unique values in `x` will be used as the x-axis values and the counts of each
        unique `x` value  will be used for as the y-axis values. If `x` and `y` are both
        provided, `x` will be used as the x-axis values and `y` will be used as the
        y-axis values.

    y : str or list, optional
        Name of a column in `data` or an iterable of values to be plotted on the
        y-axis. If `orientation` is ``"horizontal"``, and only `y` is provided, the
        unique values in `y` will be used as the y-axis values and the counts of each
        unique `y` value  will be used for as the x-axis values. If `x` and `y` are both
        provided, `x` will always be used for bar names and `y` will always be used
        for bar lengths, regardless of `orientation`.

    hue : str, optional
        Name of a column in `data` or an iterable of hue categories to be used to
        group data into stacked bars. If not provided, an un-stacked bar plot is created.

    data : pandas.DataFrame, optional
        A ``DataFrame`` object containing the input data. If provided, `x` and/or `y` should
        be column names in `data`.

    sequences : iterable of abutils.core.sequence.Sequence, optional
        An iterable of ``Sequence`` objects. If provided, `x`, `y` and/or `hue` should be annotations
        in the ``Sequence`` objects. Alternatively, `x`, `y` and/or `hue` can be an iterable of
        values to be plotted, but must be the same length as `sequences`.

    order : iterable object, optional
        List of `x` or `y` categories in the order they should be plotted. If `order` contains a
        subset of all categories found in `x` or `y`, only the supplied categories will be plotted.
        If not provided, categories will be plotted in ``natsort.natsorted()`` order.

    hue_order : iterable object, optional
        List of `hue` categories in the order they should be plotted. If `hue_order` contains a
        subset of all categories found in `hue`, only the supplied categories will be plotted.
        If not provided, `hue` categories will be plotted in ``natsort.natsorted()`` order.

    palette : dict, optional
        Dictionary mapping `hue`, `x` or `y` names to colors. If if keys in `palette` match
        more than one category, `hue` categories take priority. If `palette` is not provided,
        bars are colored using `color` (if `hue` is ``None``) or a palette is generated
        automatically using ``sns.hls_palette()``.

    color : str or iterable, optional
        Single color to be used for the bar plot. If not provided, the first color in the
        default ``Seaborn`` color palette will be used. If `highlight` is provided but
        `highlight_color` is not, `color` will be used to color highlighted bars.

    alt_color : str or iterable, default='#D3D3D3'
        Alternate color for the bar plot. Used to color categories not provided in `palette`
        or to color categories not present in `highlight`.

    orientation : str, optional
        Orientation of the plot. Options are ``'vertical'`` or ``'horizontal'``. Default is
        ``'vertical'``.

    normalize : bool, default=False
        If ``True``, normalized frequencies are plotted instead of raw counts. If multiple `hue`
        categories are present, data will be normalized such that all
        bars extend from [0,1] and each stacked bar is sized according to the `hue`'s  fraction.
        If `hue` is not provided or there is only one `hue` category, the entire
        dataset is normalized.

    highlight : iterable, optional
        List of `x` or `hue` categories to be highlighted. If `highlight_color` is provided,
        categories in `highlight` will use `highlight_color` and all others will use `alt_color`.
        If `highlight_color` is not provided, `palette` will be used. If both `highlight_color`
        and `palette` are not provided, `color` will be used.

    highlight_color : str or iterable, optional
        Color to be used for categories in `highlight`. If

    plot_kwargs : dict, optional
        Dictionary containing keyword arguments that will be passed to ``pyplot.bar()``.

    legend_kwargs : dict, optional
        Dictionary containing keyword arguments that will be passed to ``ax.legend()``.

    hide_legend : bool, default=False
        By default, a plot legend will be shown if multiple batches are plotted. If ``True``,
        the legend will not be shown.

    xlabel : str, optional
        Text for the X-axis label.

    ylabel : str, optional
        Text for the Y-axis label.

    xlabel_fontsize : int or float, default=16
        Fontsize for the X-axis label text.

    ylabel_fontsize : int or float, default=16
        Fontsize for the Y-axis label text.

    xtick_labelsize : int or float, default=14
        Fontsize for the X-axis tick labels.

    ytick_labelsize : int or float, default=14
        Fontsize for the Y-axis tick labels.

    xtick_labelrotation : int or float, default=0
        Rotation of the X-axis tick labels.

    ytick_labelrotation : int or float, default=0
        Rotation of the Y-axis tick labels.

    show :bool, default=False
        If ``True``, plot is shown and the plot ``Axes`` object is not returned. Default
        is ``False``, which does not call ``pyplot.show()`` and returns the ``Axes`` object.

    figsize : iterable object, default=[6, 4]
        List containing the figure size (as ``[x-dimension, y-dimension]``) in inches.

    figfile : str, optional
        Path at which to save the figure file. If not provided, the figure is not saved
        and is either shown (if `show` is ``True``) or the ``Axes`` object is returned.
    """
    # process input data
    if orientation.lower() == "horizontal":
        if x is None and y is not None:
            x, y = y, x
    # if data is None:
    #     _data = {}
    #     if x is not None:
    #         _data["x"] = x
    #         x = "x"
    #     if y is not None:
    #         _data["y"] = y
    #         y = "y"
    #     if hue is not None:
    #         _data["hue"] = hue
    #         hue = "hue"
    #     data = pd.DataFrame(_data)
    data, x, y, hue = process_input_data(
        x=x, y=y, hue=hue, data=data, sequences=sequences
    )

    # figure size
    if figsize is None:
        figsize = [4, 6] if orientation.lower() == "horizontal" else [6, 4]
        # if orientation == "horizontal":
        #     figsize = [4, 6]
        # else:
        #     figsize = [6, 4]

    if order is None:
        if all([isinstance(i, (int, float)) for i in data[x]]):
            x_vals = range(
                math.floor(data[x].min()), math.ceil(data[x].max()) + 1
            )  # +1 to include the last value
        else:
            x_vals = natsorted(data[x].unique())
    else:
        x_vals = order

    # process hue, if provided
    if hue is not None:
        hue_vals = data[hue]
        hue_order = hue_order if hue_order is not None else natsorted(set(hue_vals))
        hue_batches = [
            data[[_hue_val == h for _hue_val in hue_vals]] for h in hue_order
        ]
    else:
        hue_order = [
            None,
        ]
        hue_batches = [
            data,
        ]

    # process batches
    batch_data = []
    for batch in hue_batches:
        if y is not None:
            y_dict = {}
            for _x, _y in zip(batch[x], batch[y]):
                if _x in y_dict:
                    y_dict[_x] += float(_y)
                else:
                    y_dict[_x] = float(_y)
            batch_data.append(y_dict)
        else:
            y_dict = Counter(batch[x])
            batch_data.append(y_dict)
    if normalize:
        if len(batch_data) > 1:
            ytots = {xval: sum([b[xval] for b in batch_data]) for xval in x_vals}
            for y_dict in batch_data:
                for xval in x_vals:
                    yval = y_dict.get(xval, 0)
                    y_dict[xval] = yval / ytots[xval]
        else:
            for ydict in batch_data:
                tot = sum(ydict.values())
                for xval in x_vals:
                    yval = y_dict.get(xval, 0)
                    y_dict[xval] = yval / tot

    # colors
    if palette is None:
        if len(hue_batches) > 1:
            palette = {h: c for h, c in zip(hue_order, sns.hls_palette(len(hue_order)))}
        else:
            palette = {}
        # palette is a dict assigning colors to x and/or hue categories
        # if both are present, hue takes priority
    colors = []
    color = color if color is not None else sns.color_palette()[0]
    for _hue, batch in zip(hue_order, hue_batches):
        _colors = []
        for _x in x_vals:
            if highlight is not None:
                if _x in highlight:
                    if highlight_color is not None:
                        _colors.append(highlight_color)
                    else:
                        _colors.append(palette.get(_hue, palette.get(_x, color)))
                else:
                    _colors.append(alt_color)
            else:
                _colors.append(palette.get(_hue, palette.get(_x, color)))
        colors.append(_colors)

    # plot kwargs
    if orientation == "horizontal":
        default_plot_kwargs = {"height": 0.8, "linewidth": 1.5, "edgecolor": "w"}
    else:
        default_plot_kwargs = {"width": 0.8, "linewidth": 1.5, "edgecolor": "w"}
    if plot_kwargs is not None:
        default_plot_kwargs.update(plot_kwargs)
    plot_kwargs = default_plot_kwargs

    # legend kwargs
    default_legend_kwargs = {"frameon": True, "loc": "best", "fontsize": 12}
    if legend_kwargs is not None:
        default_legend_kwargs.update(legend_kwargs)
    legend_kwargs = default_legend_kwargs

    # make the plot
    if ax is None:
        plt.figure(figsize=figsize)
        ax = plt.gca()
    bottom = np.zeros(len(x_vals))
    for h, d, c in zip(hue_order, batch_data, colors):
        y_vals = np.asarray([d.get(_x, 0) for _x in x_vals])
        if orientation == "horizontal":
            ax.barh(x_vals, y_vals, left=bottom, color=c, label=h, **plot_kwargs)
        else:
            ax.bar(x_vals, y_vals, bottom=bottom, color=c, label=h, **plot_kwargs)
        bottom += y_vals

    # style the plot
    if orientation == "horizontal":
        if xlabel is None:
            xlabel = "Frequency (%)" if normalize else "Count"
    else:
        if ylabel is None:
            ylabel = "Frequency (%)" if normalize else "Count"
    ax.set_xlabel(xlabel, fontsize=xlabel_fontsize)
    ax.set_ylabel(ylabel, fontsize=ylabel_fontsize)
    ax.tick_params(
        axis="x", labelsize=xtick_labelsize, labelrotation=xtick_labelrotation
    )
    ax.tick_params(
        axis="y", labelsize=ytick_labelsize, labelrotation=ytick_labelrotation
    )
    for s in ["left", "right", "top"]:
        ax.spines[s].set_visible(False)

    # if orientation == "horizontal":
    #     ax.set_ylim([-0.75, len(x_vals) - 0.25])
    # else:
    #     ax.set_xlim([-0.75, len(x_vals) - 0.25])

    # legend
    if len(hue_batches) > 1 and not hide_legend:
        if orientation == "horizontal":
            leg_kwargs = {"loc": "center left", "bbox_to_anchor": [1.01, 0.5]}
            leg_kwargs.update(legend_kwargs)
        else:
            leg_kwargs = legend_kwargs
        ax.legend(**leg_kwargs)
    if hide_legend or palette is None:
        leg = ax.get_legend()
        if leg is not None:
            leg.remove()

    # save, show or return the ax
    if figfile is not None:
        plt.tight_layout()
        plt.savefig(figfile)
    elif show:
        plt.show()
    else:
        return ax



#!/usr/bin/env python
# filename: s3.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


# import os
# import subprocess as sp
# import sys
# import tarfile

# from . import log

# if sys.version_info[0] > 2:
#     raw_input = input


# def compress_and_upload(
#     data,
#     compressed_file,
#     s3_path,
#     multipart_chunk_size_mb=500,
#     method="gz",
#     delete=False,
#     access_key=None,
#     secret_key=None,
# ):
#     """
#     Compresses data and uploads to S3.

#     S3 upload uses ``s3cmd``, so you must either:

#         1) Manually configure ``s3cmd`` prior to use (typically using ``s3cmd --configure``).

#         2) Configure ``s3cmd`` using ``s3.configure()``.

#         3) Pass your access key and secret key to ``compress_and_upload``, which will automatically configure s3cmd.

#     .. note:

#         ``s3cmd`` configuration only needs to be done once per computer,
#         which means that relaunching a cloud instance or Docker image will
#         require re-configuration of ``s3cmd``.

#     Args:

#         data: Can be one of three things:

#             1) Path to a single file

#             2) Path to a directory

#             3) A list of one or more paths to files or directories

#         compressed_file (str): Path to the compressed file. Required.

#         s3_path (str): The S3 path, with the filename omitted. The S3 filename
#           will be the basename of the ``compressed_file``. For example::

#             compress_and_upload(data='/path/to/data',
#                                 compressed_file='/path/to/compressed.tar.gz',
#                                 s3_path='s3://my_bucket/path/to/')

#           will result in an uploaded S3 path of ``s3://my_bucket/path/to/compressed.tar.gz``

#         method (str): Compression method. Options are ``'gz'`` (gzip) or ``'bz2'`` (bzip2).
#             Default is ``'gz'``.

#         delete (bool): If ``True``, the ``compressed_file`` will be deleted after upload
#             to S3. Default is ``False``.

#         access_key (str): AWS access key.

#         secret_key (str): AWS secret key.
#     """
#     logger = log.get_logger("s3")
#     if all([access_key, secret_key]):
#         configure(access_key=access_key, secret_key=secret_key, logger=logger)
#     compress(data, compressed_file, fmt=method, logger=logger)
#     put(
#         compressed_file,
#         s3_path,
#         multipart_chunk_size_mb=multipart_chunk_size_mb,
#         logger=logger,
#     )
#     if delete:
#         os.unlink(compressed_file)


# def put(f, s3_path, multipart_chunk_size_mb=500, logger=None):
#     """
#     Uploads a single file to S3, using s3cmd.

#     Args:

#         f (str): Path to a single file.

#         s3_path (str): The S3 path, with the filename omitted. The S3 filename
#             will be the basename of the ``f``. For example::

#                 put(f='/path/to/myfile.tar.gz', s3_path='s3://my_bucket/path/to/')

#             will result in an uploaded S3 path of ``s3://my_bucket/path/to/myfile.tar.gz``
#     """
#     if not logger:
#         logger = log.get_logger("s3")
#     fname = os.path.basename(f)
#     target = os.path.join(s3_path, fname)
#     s3cmd_cline = "s3cmd put {} {} --multipart-chunk-size-mb {}".format(
#         f, target, multipart_chunk_size_mb
#     )
#     print_put_info(fname, target, logger)
#     s3cmd = sp.Popen(s3cmd_cline, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
#     stdout, stderr = s3cmd.communicate()


# def print_put_info(fname, target, logger):
#     logger.info("")
#     logger.info("")
#     logger.info("")
#     logger.info("-" * 25)
#     logger.info("UPLOADING TO S3")
#     logger.info("-" * 25)
#     logger.info("")
#     logger.info("File: {}".format(fname))
#     logger.info("Target S3 location: {}".format(target))


# def compress(d, output, fmt="gz", logger=None):
#     """
#     Creates a compressed/uncompressed tar file.

#     Args:

#         d: Can be one of three things:

#             1. the path to a single file, as a string

#             2. the path to a single directory, as a string

#             3. an iterable of file or directory paths

#         output (str): Output file path.

#         fmt: Compression method. Options are ``'gz'`` (gzip),
#             ``'bz2'`` (bzip2) and ``'none'`` (uncompressed). Default is ``'gz'``.
#     """
#     if not logger:
#         logger = log.get_logger("s3")
#     if type(d) not in [list, tuple]:
#         d = [
#             d,
#         ]
#     d = [os.path.expanduser(_d) for _d in d]
#     print_compress_info(d, output, compress, logger)
#     if fmt.lower() == "none":
#         fmt = ""
#     elif fmt.lower() not in ["gz", "bz2"]:
#         logger.info(
#             'Compression option ("{}") is invalid.\nFalling back to uncompressed.'.format(
#                 fmt
#             )
#         )
#         fmt = ""
#     output = os.path.expanduser(output)
#     tar = tarfile.open(output, "w:{}".format(fmt))
#     for obj in d:
#         tar.add(obj)
#     tar.close()
#     return output


# def print_compress_info(d, output, compress, logger):
#     if not logger:
#         logger = log.get_logger("s3")
#     dirs = [obj for obj in d if os.path.isdir(obj)]
#     files = [obj for obj in d if os.path.isfile(obj)]
#     logger.info("")
#     logger.info("")
#     logger.info("")
#     logger.info("-" * 25)
#     logger.info("COMPRESSING DATA")
#     logger.info("-" * 25)
#     logger.info("")
#     logger.info("Ouptut file: {}".format(output))
#     logger.info("Compression: {}".format(compress.lower()))
#     if dirs:
#         d = "directories" if len(dirs) > 1 else "directory"
#         logger.info("Found {} {} to compress: {}".format(len(dirs), d, ", ".join(dirs)))
#     if files:
#         f = "files" if len(files) > 1 else "file"
#         logger.info(
#             "Found {} {} to compress: {}".format(len(files), f, ", ".join(files))
#         )


# def configure(access_key=None, secret_key=None, logger=None):
#     """
#     Configures s3cmd prior to first use.

#     If no arguments are provided, you will be prompted to enter
#     the access key and secret key interactively.

#     Args:

#         access_key (str): AWS access key

#         secret_key (str): AWS secret key
#     """
#     if not logger:
#         logger = log.get_logger("s3")
#     if not all([access_key, secret_key]):
#         logger.info("")
#         access_key = input("AWS Access Key: ")
#         secret_key = input("AWS Secret Key: ")
#     _write_config(access_key, secret_key)
#     logger.info("")
#     logger.info("Completed writing S3 config file.")
#     logger.info("")


# def _write_config(access_key, secret_key):
#     cfg_string = "[default]\n"
#     cfg_string += "access_key = {}\n".format(access_key)
#     cfg_string += "secret_key = {}\n".format(secret_key)
#     cfg_string += CONFIG_DEFAULTS
#     cfg_file = os.path.expanduser("~/.s3cfg")
#     open(cfg_file, "w").write(cfg_string)


# CONFIG_DEFAULTS = """
# access_token =
# add_encoding_exts =
# add_headers =
# bucket_location = US
# cache_file =
# cloudfront_host = cloudfront.amazonaws.com
# default_mime_type = binary/octet-stream
# delay_updates = False
# delete_after = False
# delete_after_fetch = False
# delete_removed = False
# dry_run = False
# enable_multipart = True
# encoding = UTF-8
# encrypt = False
# expiry_date =
# expiry_days =
# expiry_prefix =
# follow_symlinks = False
# force = False
# get_continue = False
# gpg_command = /usr/bin/gpg
# gpg_decrypt = %(gpg_command)s -d --verbose --no-use-agent --batch --yes --passphrase-fd %(passphrase_fd)s -o %(output_file)s %(input_file)s
# gpg_encrypt = %(gpg_command)s -c --verbose --no-use-agent --batch --yes --passphrase-fd %(passphrase_fd)s -o %(output_file)s %(input_file)s
# gpg_passphrase =
# guess_mime_type = True
# host_base = s3.amazonaws.com
# host_bucket = %(bucket)s.s3.amazonaws.com
# human_readable_sizes = False
# ignore_failed_copy = False
# invalidate_default_index_on_cf = False
# invalidate_default_index_root_on_cf = True
# invalidate_on_cf = False
# list_md5 = False
# log_target_prefix =
# max_delete = -1
# mime_type =
# multipart_chunk_size_mb = 250
# preserve_attrs = True
# progress_meter = True
# proxy_host =
# proxy_port = 0
# put_continue = False
# recursive = False
# recv_chunk = 4096
# reduced_redundancy = False
# restore_days = 1
# send_chunk = 4096
# server_side_encryption = False
# simpledb_host = sdb.amazonaws.com
# skip_existing = False
# socket_timeout = 300
# urlencoding_mode = normal
# use_https = True
# use_mime_magic = True
# verbosity = WARNING
# website_endpoint = http://%(bucket)s.s3-website-% (location)s.amazonaws.com/
# website_error =
# website_index = index.html
# """



#!/usr/bin/env python
# filename: io.py


#
# Copyright (c) 2020 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import glob
import gzip
import os
import re
import tempfile
from typing import Iterable, Optional, Union

import pandas as pd
import polars as pl
from Bio.SeqRecord import SeqRecord

# import pyarrow as pa
# import pyarrow.parquet as pq
from natsort import natsorted

from .core.pair import (
    Pair,
    # pairs_from_pandas,
    pairs_from_polars,
    # pairs_to_csv,
    # pairs_to_pandas,
    # pairs_to_parquet,
    pairs_to_polars,
)
from .core.sequence import (
    Sequence,
    determine_fastx_format,
    from_mongodb,
    parse_fasta,
    parse_fastq,
    parse_fastx,
    # read_airr,
    # read_csv,
    read_fasta,
    read_fastq,
    read_fastx,
    read_json,
    # read_parquet,
    # sequences_from_pandas,
    sequences_from_polars,
    # sequences_to_csv,
    # sequences_to_pandas,
    # sequences_to_parquet,
    sequences_to_polars,
    # to_airr,
    # to_fasta,
    to_fastq,
)
from .utils.path import delete_files, list_files, make_dir, rename_file

# =======================
#   General file I/O
# =======================


def make_dir(directory: str) -> None:
    """
    Makes a directory, if it doesn't already exist.

    Parameters
    ----------
    directory : str
        Path to a directory.

    """
    if not os.path.exists(directory):
        os.makedirs(os.path.abspath(directory))


def list_files(
    directory: str,
    extension: Union[str, Iterable, None] = None,
    recursive: bool = False,
    match: Optional[str] = None,
    ignore_dot_files: bool = True,
) -> Iterable[str]:
    """
    Lists files in a given directory.

    Parameters
    ----------
    directory : str
        Path to a directory. If a file path is passed instead, the returned list of files will contain
        only that file path.

    extension : str
        If supplied, only files that end with the specificied extension(s) will be returned. Can be either
        a string or a list of strings. Extension evaluation is case-insensitive and can match complex
        extensions (e.g. '.fastq.gz'). Default is ``None``, which returns all files in the directory,
        regardless of extension.

    recursive : bool, default=False
        If ``True``, the directory will be searched recursively, and all files in all subdirectories will be returned.

    match : str, optional
        If supplied, only files that match the specified pattern will be returned. Regular expressions are supported.

    ignore_dot_files : bool, default=True
        If ``True``, dot files (hidden files) will be ignored.

    Returns
    -------
    Iterable[str]

    """
    directory = os.path.abspath(directory)
    if os.path.exists(directory):
        if os.path.isdir(directory):
            if recursive:
                files = []
                for root, dirs, _files in os.walk(directory):
                    for f in _files:
                        file_path = os.path.join(root, f)
                        files.append(file_path)
            else:
                files = natsorted(glob.glob(directory + "/*"))
        else:
            files = [directory]
    else:
        raise ValueError(f"Directory {directory} does not exist.")
    if extension is not None:
        if isinstance(extension, str):
            extension = [
                extension,
            ]
        files = [
            f
            for f in files
            if any(
                [
                    any([f.lower().endswith(e.lower()) for e in extension]),
                    any([f.endswith(e.upper()) for e in extension]),
                    any([f.endswith(e.lower()) for e in extension]),
                ]
            )
        ]
    if ignore_dot_files:
        files = [f for f in files if not os.path.basename(f).startswith(".")]
    if match is not None:
        files = [f for f in files if re.match(match, os.path.basename(f)) is not None]
    return files


def rename_file(file: str, new_name: str) -> None:
    """
    Renames a file.

    Parameters
    ----------
    file : str
        Path to the file to be renamed.

    new_name : str
        New name for the file.

    """
    os.rename(file, new_name)


def delete_files(files: Union[str, Iterable]) -> None:
    """
    Deletes files.

    Parameters
    ----------
    files : Union[str, Iterable]
        Path to a file or an iterable of file paths.

    """
    if isinstance(files, str):
        files = [
            files,
        ]
    for f in files:
        if os.path.exists(f):
            os.remove(f)


# ============================================
#      File splitting and concatenation
# ============================================


def concatenate_files(files: Iterable[str], output_file: str) -> None:
    """
    Concatenates multiple files into a single file.

    Parameters
    ----------
    files : Iterable[str]
        Iterable of file paths.

    output_file : str
        Path to the output file.

    """
    with open(output_file, "w") as outfile:
        for fname in files:
            with open(fname) as infile:
                for line in infile:
                    outfile.write(line)


def split_parquet(
    parquet_file: str,
    output_directory: str,
    num_rows: int = 500,
    num_splits: Optional[int] = None,
    split_prefix: str = "chunk_",
    start_numbering_at: int = 0,
) -> Iterable[str]:
    """
    Splits a parquet file into multiple files.

    Parameters
    ----------
    parquet_file : str
        Path to the parquet file to be split.

    output_directory : str
        Path to the directory where the split files will be saved. If the directory does not exist, it will be created.

    num_rows : int, optional
        Number of rows per split file. Default is 500. If ``num_splits`` is supplied, this argument is ignored.

    num_splits : int, optional
        Number of split files to create. If not supplied, ``num_rows`` is used to determine the number of split files.

    split_prefix : str, optional
        Prefix for the split files, which is followed directly by the file number. Default is "chunk_".

    start_numbering_at : int, optional
        Start numbering the split files at this number. Default is 0.

    Returns
    -------
    Iterable[str]
        Iterable of file paths for the split files.

    """
    import pyarrow as pa  # for some reason, pyarrow breaks readthedocs if imported up top
    import pyarrow.parquet as pq

    # outputs
    output_files = []
    output_directory = os.path.abspath(output_directory)
    make_dir(output_directory)

    # read parquet file
    pqfile = pq.ParquetFile(parquet_file)

    # number of rows per split file
    if num_splits is not None:
        total_rows = pqfile.metadata.num_rows
        num_rows = total_rows // num_splits
        if total_rows % num_splits != 0:
            num_rows += 1

    # split into batches and process
    batches = pqfile.iter_batches(batch_size=num_rows)
    for i, batch in enumerate(batches, start=start_numbering_at):
        table = pa.Table.from_batches([batch])
        output_file = os.path.join(output_directory, f"{split_prefix}{i}.parquet")
        pq.write_table(table, output_file)
        output_files.append(output_file)
    return output_files


def concatenate_parquet(parquet_files: Iterable[str], output_file: str) -> str:
    """
    Concatenates multiple parquet files into a single file.

    Parameters
    ----------
    parquet_files : Iterable[str]
        Iterable of file paths to the parquet files to be concatenated.

    output_file : str
        Path to the output file.

    Returns
    -------
    str
        Path to the output file.

    """
    make_dir(os.path.dirname(output_file))

    # do all parquet operations lazily to save memory
    dfs = [pl.scan_parquet(f) for f in parquet_files]
    df = pl.concat(dfs)
    df.sink_parquet(output_file)
    return output_file


def split_csv(
    csv_file: str, output_directory: str, separator: str = ",", chunksize: int = 500
) -> Iterable[str]:
    """
    Splits a CSV file into multiple files, each containing a specified number of rows, including the header in each split.

    Parameters
    ----------
    csv_file : str
        Path to the CSV file to be split.

    output_directory : str
        Path to the directory where the split files will be saved. If the directory does not exist, it will be created.

    separator : str, default=","
        Column separator. Default is `","`.

    chunksize : int, default=500
        Number of rows per split file. Default is 500. The last file may contain fewer rows than this number.

    Returns
    -------
    Iterable[str]
        Iterable of file paths for the split files.

    """
    make_dir(output_directory)

    output_files = []
    output_basename = os.path.splitext(os.path.basename(csv_file))[0]

    for i, chunk in enumerate(
        pd.read_csv(csv_file, sep=separator, chunksize=chunksize)
    ):
        output_file = os.path.join(output_directory, f"{output_basename}_part_{i}.csv")
        chunk.to_csv(output_file, sep=separator, index=False, mode="w", header=True)
        output_files.append(output_file)

    return output_files


def concatenate_csv(
    csv_files: Iterable[str], output_file: str, separator: str = ","
) -> str:
    """
    Concatenates multiple CSV files into a single file.

    Parameters
    ----------
    csv_files : Iterable[str]
        Iterable of file paths to the CSV files to be concatenated.

    output_file : str
        Path to the output file.

    separator : str, default=","
        Column separator. Default is ``","``.

    Returns
    -------
    str
        Path to the output file.

    """
    make_dir(os.path.dirname(output_file))

    # do all csv operations lazily to save memory
    dfs = [pl.scan_csv(f, separator=separator) for f in csv_files]
    df = pl.concat(dfs)
    df.sink_csv(output_file, separator=separator)
    return output_file


def split_airr(
    airr_file: str, output_directory: str, chunksize: int = 500
) -> Iterable[str]:
    """
    Splits an AIRR-formatted file into multiple files.

    Parameters
    ----------
    airr_file : str
        Path to the AIRR-formatted file to be split.

    output_directory : str
        Path to the directory where the split files will be saved.

    chunksize : int, default=500
        Number of rows per split file. Default is 500. The last file may contain fewer rows than this number.

    Returns
    -------
    Iterable[str]
        Iterable of file paths for the split files.

    """
    return split_csv(airr_file, output_directory, separator="\t", chunksize=chunksize)


def concatenate_airr(airr_files: Iterable[str], output_file: str) -> str:
    """
    Concatenates multiple AIRR-formatted files into a single file.

    Parameters
    ----------
    airr_files : Iterable[str]
        Iterable of file paths to the AIRR-formatted files to be concatenated.

    output_file : str
        Path to the output file.

    Returns
    -------
    str
        Path to the output file.

    """
    return concatenate_csv(airr_files, output_file, separator="\t")


def split_fastx(
    fastx_file: str,
    output_directory: str,
    chunksize: int = 500,
    start_numbering_at: int = 0,
    fmt: Optional[str] = None,
) -> Iterable[str]:
    """
    Splits a FASTA or FASTQ file into multiple files, each containing a specified number of sequences.

    Parameters
    ----------
    fastx_file : str
        Path to the FASTA or FASTQ file to be split.

    output_directory : str
        Path to the directory where the split files will be saved. If the directory does not exist, it will be created.

    chunksize : int, optional
        Number of sequences per split file. Default is 500. The last file may contain fewer sequences than this number.

    start_numbering_at : int, optional
        Start numbering the split files at this number. Default is 0.

    fmt : str, optional
        Format of the input file. If not supplied, the format will be determined automatically.

    Returns
    -------
    Iterable[str]
        Iterable of file paths for the split files.

    """
    # outputs
    output_files = []
    output_directory = os.path.abspath(output_directory)
    make_dir(output_directory)
    output_basename = ".".join(
        os.path.basename(fastx_file).rstrip(".gz").split(".")[:-1]
    )

    # get input file format
    if fmt is None:
        fmt = determine_fastx_format(fastx_file)

    # split sequences into chunks, write the chunks to files
    file_num = start_numbering_at
    chunk = []
    for s in parse_fastx(fastx_file):
        chunk.append(s.fastq if fmt == "fastq" else s.fasta)
        if len(chunk) == chunksize:
            output_file = os.path.join(
                output_directory, f"{output_basename}_{file_num}.{fmt}"
            )
            with open(output_file, "w") as f:
                f.write("\n".join(chunk))
            chunk = []
            file_num += 1
            output_files.append(output_file)

    # write the last chunk to a file
    if chunk:
        output_file = os.path.join(
            output_directory, f"{output_basename}_{file_num}.{fmt}"
        )
        with open(output_file, "w") as f:
            f.write("\n".join(chunk))
        output_files.append(output_file)

    return output_files


def split_fasta(
    fasta_file: str,
    output_directory: str,
    chunksize: int = 500,
    start_numbering_at: int = 0,
) -> Iterable[str]:
    """
    Splits a FASTA or FASTQ file into multiple files, each containing a specified number of sequences.

    Parameters
    ----------
    fasta_file : str
        Path to the FASTA file to be split.

    output_directory : str
        Path to the directory where the split files will be saved. If the directory does not exist, it will be created.

    chunksize : int, optional
        Number of sequences per split file. Default is 500. The last file may contain fewer sequences than this number.

    start_numbering_at : int, optional
        Start numbering the split files at this number. Default is 0.

    Returns
    -------
    Iterable[str]
        Iterable of file paths for the split files.

    """
    return split_fastx(
        fastx_file=fasta_file,
        output_directory=output_directory,
        chunksize=chunksize,
        start_numbering_at=start_numbering_at,
        fmt="fasta",
    )


def split_fastq(
    fastq_file: str,
    output_directory: str,
    chunksize: int = 500,
    start_numbering_at: int = 0,
) -> Iterable[str]:
    """
    Splits a FASTA or FASTQ file into multiple files, each containing a specified number of sequences.

    Parameters
    ----------
    fastq_file : str
        Path to the FASTQ file to be split.

    output_directory : str
        Path to the directory where the split files will be saved. If the directory does not exist, it will be created.

    chunksize : int, optional
        Number of sequences per split file. Default is 500. The last file may contain fewer sequences than this number.

    start_numbering_at : int, optional
        Start numbering the split files at this number. Default is 0.

    Returns
    -------
    Iterable[str]
        Iterable of file paths for the split files.

    """
    return split_fastx(
        fastx_file=fastq_file,
        output_directory=output_directory,
        chunksize=chunksize,
        start_numbering_at=start_numbering_at,
        fmt="fastq",
    )


# ===============================
#     Sequence and Pair I/O
# ===============================


def to_fasta(
    sequences: Union[str, Iterable],
    fasta_file: Optional[str] = None,
    id_key: Optional[str] = None,
    sequence_key: Optional[str] = None,
    tempfile_dir: Optional[str] = None,
    append_chain: bool = True,
    as_string: bool = False,
) -> str:
    """
    Writes sequences to a FASTA-formatted file or returns a FASTA-formatted string.

    Parameters
    ----------
    sequences : str or Iterable
        Accepts any of the following:
            1. list of abutils ``Sequence`` and/or ``Pair`` objects
            2. FASTA/Q-formatted string
            3. path to a FASTA/Q-formatted file
            4. list of BioPython ``SeqRecord`` objects
            5. list of lists/tuples, of the format ``[sequence_id, sequence]``
        Required.

        .. note::
            Processing a list containing a mixture of ``Sequence`` and/or ``Pair`` objects is supported.

    fasta_file : str, default=None
        Path to the output FASTA file. If neither `fasta_file` nor `tempfile_dir`
        are provided, a FASTA-formatted string will be returned.

    id_key : str, default=None
        Name of the annotation field containing the sequence ID. If not provided,
        ``sequence.id`` is used.

    sequence_key : str, default=None
        Name of the annotation field containg the sequence. If not provided,
        ``sequence.sequence`` is used.

    tempfile_dir : str, optional
        If `fasta_file` is not provided, directory into which the tempfile
        should be created. If the directory does not exist, it will be
        created.

    append_chain : bool, default=True
        If ``True``, the chain (heavy or light) will be appended to the sequence name:
        ``>MySequence_heavy``.

        .. note::
            This option is ignored unless a list containing ``Pair`` objects is provided.

    as_string : bool, default=False
        Deprecated. Kept for backwards compatibility.


    Returns
    --------
    fasta : str
        Path to a FASTA file or a FASTA-formatted string

    """
    if isinstance(sequences, str):
        # if sequences is already a FASTA/Q-formatted file
        if os.path.isfile(sequences):
            fasta_string = "\n".join([s.fasta for s in parse_fasta(sequences)])
        # if sequences is a FASTA-formatted string
        else:
            fasta_string = sequences
    # if sequences is a list of Sequences
    elif all([isinstance(s, (Sequence, Pair)) for s in sequences]):
        fasta_strings = []
        for s in sequences:
            if isinstance(s, Pair):
                fasta_strings.append(
                    s.fasta(
                        name_field=id_key,
                        sequence_field=sequence_key,
                        append_chain=append_chain,
                    )
                )
            else:
                fasta_strings.append(
                    s.as_fasta(name_field=id_key, seq_field=sequence_key)
                )
        fasta_string = "\n".join(fasta_strings)
    # anything else..
    else:
        fasta_string = "\n".join(
            [Sequence(s, id_key=id_key, seq_key=sequence_key).fasta for s in sequences]
        )
    # output
    if fasta_file is not None:
        make_dir(os.path.dirname(fasta_file))
        with open(fasta_file, "w") as f:
            f.write(fasta_string)
        return fasta_file
    elif tempfile_dir is not None:
        make_dir(tempfile_dir)
        ff = tempfile.NamedTemporaryFile(dir=tempfile_dir, delete=False)
        ff.close()
        with open(ff.name, "w") as f:
            f.write(fasta_string)
        return ff.name
    return fasta_string


def from_polars(
    df: Union[pl.DataFrame, pl.LazyFrame],
    match: Optional[dict] = None,
    fields: Optional[Iterable] = None,
    id_key: str = "sequence_id",
    sequence_key: str = "sequence",
) -> Iterable[Union[Sequence, Pair]]:
    """
    Reads a Polars DataFrame and returns a list of ``Sequence`` or ``Pair`` objects.

    Parameters
    ----------
    df : polars.DataFrame or polars.LazyFrame
        The input Polars DataFrame or LazyFrame.

    match : dict, optional
        A ``dict`` for filtering sequences from the input file.
        Sequences must match all conditions to be returned. For example,
        the following ``dict`` will filter out all sequences for which
        the ``'v_gene:0'`` field is not ``'IGHV1-2'``:

        .. code-block:: python

            {'v_gene:0': 'IGHV1-2'}

    fields : list, optional
        A list of fields to be read from the input file. If not provided, all fields will be read.

    id_key : str, default="sequence_id"
        The name of the column that contains the sequence IDs. Default is ``"sequence_id"``.

    sequence_key : str, default="sequence"
        The name of the column that contains the sequence data. Default is ``"sequence"``.

    Returns
    -------
    Iterable[Union[Sequence, Pair]]
        A list of ``Sequence`` or ``Pair`` objects.

    """
    if isinstance(df, pl.LazyFrame):
        df = df.collect()
    if any([c in df.columns for c in ["sequence:0", "sequence_id:0"]]):
        return pairs_from_polars(
            df, match=match, fields=fields, id_key=id_key, sequence_key=sequence_key
        )
    else:
        return sequences_from_polars(
            df, match=match, fields=fields, id_key=id_key, sequence_key=sequence_key
        )


def to_polars(
    sequences: Iterable[Union[Sequence, Pair]],
    annotations: Optional[Iterable[str]] = None,
    columns: Optional[Iterable] = None,
    properties: Optional[Iterable[str]] = None,
    sequence_properties: Optional[Iterable[str]] = None,
    drop_na_columns: bool = True,
    order: Optional[Iterable[str]] = None,
    exclude: Optional[Union[str, Iterable[str]]] = None,
    leading: Optional[Union[str, Iterable[str]]] = None,
) -> Optional[pl.DataFrame]:
    """
    Saves a list of ``Pair`` objects to a Polars DataFrame.

    Parameters
    ----------
    sequences : Iterable[Sequence, Pair]
        List of ``Sequence`` or ``Pair`` objects to be saved to a Polars DataFrame. Required.

    annotations : list, default=None
        A list of annotation fields to be included in the Polars DataFrame. For ``Sequence``
        objects, this refers to fields in the ``annotations`` field. For ``Pair`` objects,
        this refers to fields in the heavy and light chain annotations.

    columns : list, default=None
        Used only for ``Pair`` objects. A list of fields to be retained in the output Polars
        DataFrame. Fields should be column names in the input file, such as ``"sequence:0"``,
        ``"sequence:1"``, ``"name"``, etc. This option is provided to allow differential
        selection of heavy and light chain fields. For cases in which the same fields will
        be selected for both chains, it is recommended to use ``annotations`` instead.

    properties : list, default=None
        A list of properties to be included in the Polars DataFrame. If not provided, everything
        in the ``annotations`` field of each heavy/light chain will be included.

    sequence_properties : list, default=None
        A list of sequence properties to be included. Differs from ``properties``, which
        refers to properties of the ``Pair`` object. These properties are those of the
        heavy/light ``Sequence`` objects. Ignored if the input is a list of ``Sequence`` objects.

    drop_na_columns : bool, default=True
        If ``True``, columns with all ``NaN`` values will be dropped from the Polars DataFrame.
        Default is ``True``.

    order : list, default=None
        A list of fields in the order they should appear in the Polars DataFrame.

    exclude : str or list, default=None
        Field or list of fields to be excluded from the Polars DataFrame.

    leading : str or list, default=None
        Field or list of fields to appear first in the Polars DataFrame. Supercedes ``order``, so
        if both are provided, fields in ``leading`` will appear first in the Polars DataFrame and
        remaining fields will appear in the order provided in ``order``.

    Returns
    -------
    pl.DataFrame
        A ``polars.DataFrame`` object.

    """
    if all([isinstance(s, Sequence) for s in sequences]):
        return sequences_to_polars(
            sequences,
            annotations=annotations,
            properties=properties,
            drop_na_columns=drop_na_columns,
            order=order,
            exclude=exclude,
            leading=leading,
        )
    elif all([isinstance(s, Pair) for s in sequences]):
        return pairs_to_polars(
            sequences,
            annotations=annotations,
            columns=columns,
            properties=properties,
            sequence_properties=sequence_properties,
            drop_na_columns=drop_na_columns,
            order=order,
            exclude=exclude,
            leading=leading,
        )
    else:
        raise ValueError(
            "All elements in the input list must be of the same type (either Sequence or Pair)."
        )


def from_pandas(
    df: pd.DataFrame,
    match: Optional[dict] = None,
    fields: Optional[Iterable] = None,
    id_key: str = "sequence_id",
    sequence_key: str = "sequence",
) -> Iterable[Union[Sequence, Pair]]:
    """
    Reads a Pandas DataFrame and returns a list of ``Sequence`` or ``Pair`` objects.

    Parameters
    ----------
    df : pd.DataFrame
        The input Pandas DataFrame.

    match : dict, optional
        A ``dict`` for filtering sequences from the input file.
        Sequences must match all conditions to be returned. For example,
        the following ``dict`` will filter out all sequences for which
        the ``'v_gene:0'`` field is not ``'IGHV1-2'``:

        .. code-block:: python

            {'v_gene:0': 'IGHV1-2'}

    fields : list, optional
        A list of fields to be read from the input file. If not provided, all fields will be read.

    id_key : str, default="sequence_id"
        The name of the column that contains the sequence IDs. Default is ``"sequence_id"``.

    sequence_key : str, default="sequence"
        The name of the column that contains the sequence data. Default is ``"sequence"``.

    Returns
    -------
    Iterable[Union[Sequence, Pair]]
        A list of ``Sequence`` or ``Pair`` objects.

    """
    df = pl.from_pandas(df)
    return from_polars(
        df, match=match, fields=fields, id_key=id_key, sequence_key=sequence_key
    )


def to_pandas(
    sequences: Iterable[Union[Sequence, Pair]],
    annotations: Optional[Iterable[str]] = None,
    columns: Optional[Iterable] = None,
    properties: Optional[Iterable[str]] = None,
    sequence_properties: Optional[Iterable[str]] = None,
    drop_na_columns: bool = True,
    order: Optional[Iterable[str]] = None,
    exclude: Optional[Union[str, Iterable[str]]] = None,
    leading: Optional[Union[str, Iterable[str]]] = None,
) -> Optional[pl.DataFrame]:
    """
    Saves a list of ``Pair`` objects to a Pandas DataFrame.

    Parameters
    ----------
    sequences : Iterable[Sequence, Pair]
        List of ``Sequence`` or ``Pair`` objects to be saved to a Pandas DataFrame. Required.

    annotations : list, default=None
        A list of annotation fields to be included in the Pandas DataFrame. For ``Sequence``
        objects, this refers to fields in the ``annotations`` field. For ``Pair`` objects,
        this refers to fields in the heavy and light chain annotations.

    columns : list, default=None
        Used only for ``Pair`` objects. A list of fields to be retained in the output Pandas
        DataFrame. Fields should be column names in the input file, such as ``"sequence:0"``,
        ``"sequence:1"``, ``"name"``, etc. This option is provided to allow differential
        selection of heavy and light chain fields. For cases in which the same fields will
        be selected for both chains, it is recommended to use ``annotations`` instead.

    properties : list, default=None
        A list of properties to be included in the Pandas DataFrame. If not provided, everything
        in the ``annotations`` field of each heavy/light chain will be included.

    sequence_properties : list, default=None
        A list of sequence properties to be included. Differs from ``properties``, which
        refers to properties of the ``Pair`` object. These properties are those of the
        heavy/light ``Sequence`` objects. Ignored if the input is a list of ``Sequence`` objects.

    drop_na_columns : bool, default=True
        If ``True``, columns with all ``NaN`` values will be dropped from the Pandas DataFrame.
        Default is ``True``.

    order : list, default=None
        A list of fields in the order they should appear in the Pandas DataFrame.

    exclude : str or list, default=None
        Field or list of fields to be excluded from the Pandas DataFrame.

    leading : str or list, default=None
        Field or list of fields to appear first in the Pandas DataFrame. Supercedes ``order``, so
        if both are provided, fields in ``leading`` will appear first in the Pandas DataFrame and
        remaining fields will appear in the order provided in ``order``.

    Returns
    -------
    pl.DataFrame
        A ``polars.DataFrame`` object.

    """
    if all([isinstance(s, Sequence) for s in sequences]):
        df = sequences_to_polars(
            sequences,
            annotations=annotations,
            columns=columns,
            properties=properties,
            drop_na_columns=drop_na_columns,
            order=order,
            exclude=exclude,
            leading=leading,
        )
    elif all([isinstance(s, Pair) for s in sequences]):
        df = pairs_to_polars(
            sequences,
            annotations=annotations,
            columns=columns,
            properties=properties,
            sequence_properties=sequence_properties,
            drop_na_columns=drop_na_columns,
            order=order,
            exclude=exclude,
            leading=leading,
        )
    else:
        raise ValueError(
            "All elements in the input list must be of the same type (either Sequence or Pair)."
        )

    return df.to_pandas()


def read_csv(
    csv_file: str,
    separator: str = ",",
    match: Optional[dict] = None,
    fields: Optional[Iterable] = None,
    id_key: str = "sequence_id",
    sequence_key: str = "sequence",
    as_dataframe: bool = False,
) -> Iterable[Union[Sequence, Pair]]:
    """
    Reads a CSV file and returns a list of ``Sequence`` or ``Pair`` objects.

    Parameters
    ----------
    csv_file : str
        Path to the CSV file to be read.

    separator : str, default=","
        Column separator. Default is ``","``.

    match : dict, optional
        A ``dict`` for filtering sequences from the input file.
        Sequences must match all conditions to be returned. For example,
        the following ``dict`` will filter out all sequences for which
        the ``'v_gene:0'`` field is not ``'IGHV1-2'``:

        .. code-block:: python

            {'v_gene:0': 'IGHV1-2'}

    fields : list, optional
        A list of fields to be read from the input file. If not provided, all fields will be read.

    id_key : str, default="sequence_id"
        The name of the column that contains the sequence IDs. Default is ``"sequence_id"``.

    sequence_key : str, default="sequence"
        The name of the column that contains the sequence data. Default is ``"sequence"``.

    as_dataframe : bool, default=False
        If ``True``, the function will return a ``polars.DataFrame`` object.

        .. note::
            If ``True``, ``fields`` will be used to select columns from the input file, but
            ``match`` will be ignored.

    Returns
    -------
    Iterable[Union[Sequence, Pair]] or polars.DataFrame
        A list of ``Sequence`` or ``Pair`` objects or a ``polars.DataFrame`` object.

    """
    df = pl.read_csv(csv_file, separator=separator, infer_schema_length=None)
    if as_dataframe:
        if fields is not None:
            df = df.select(fields)
        return df
    if any([c in df.columns for c in ["sequence:0", "sequence_id:0"]]):
        return pairs_from_polars(
            df, match=match, fields=fields, id_key=id_key, sequence_key=sequence_key
        )
    else:
        return sequences_from_polars(
            df, match=match, fields=fields, id_key=id_key, sequence_key=sequence_key
        )


def to_csv(
    sequences: Iterable[Union[Sequence, Pair]],
    csv_file: str,
    separator: str = ",",
    header: bool = True,
    annotations: Optional[Iterable[str]] = None,
    columns: Optional[Iterable] = None,
    properties: Optional[Iterable[str]] = None,
    sequence_properties: Optional[Iterable[str]] = None,
    drop_na_columns: bool = True,
    order: Optional[Iterable[str]] = None,
    exclude: Optional[Union[str, Iterable[str]]] = None,
    leading: Optional[Union[str, Iterable[str]]] = None,
) -> None:
    """
    Saves a list of ``Pair`` objects to a CSV file.

    Parameters
    ----------
    sequences : Iterable[Sequence, Pair]
        List of ``Sequence`` or ``Pair`` objects to be saved to a CSV file. Required.

    csv_file : str
        Path to the output CSV file. Required.

    separator : str, default=","
        Column separator. Default is ``","``.

    header : bool, default=True
        If ``True``, the CSV file will contain a header row. Default is ``True``.

    annotations : list, default=None
        A list of annotation fields to be included in the CSV file. For ``Sequence``
        objects, this refers to fields in the ``annotations`` field. For ``Pair`` objects,
        this refers to fields in the heavy and light chain annotations.

    columns : list, default=None
        Used only for ``Pair`` objects. A list of fields to be retained in the output CSV
        file. Fields should be column names in the input file, such as ``"sequence:0"``,
        ``"sequence:1"``, ``"name"``, etc. This option is provided to allow differential
        selection of heavy and light chain fields. For cases in which the same fields will
        be selected for both chains, it is recommended to use ``annotations`` instead.

    properties : list, default=None
        A list of properties to be included in the CSV file. If not provided, everything
        in the ``annotations`` field of each heavy/light chain will be included.

    sequence_properties : list, default=None
        A list of sequence properties to be included. Differs from ``properties``, which
        refers to properties of the ``Pair`` object. These properties are those of the
        heavy/light ``Sequence`` objects. Ignored if the input is a list of ``Sequence`` objects.

    drop_na_columns : bool, default=True
        If ``True``, columns with all ``NaN`` values will be dropped from the CSV file.
        Default is ``True``.

    order : list, default=None
        A list of fields in the order they should appear in the CSV file.

    exclude : str or list, default=None
        Field or list of fields to be excluded from the CSV file.

    leading : str or list, default=None
        Field or list of fields to appear first in the CSV file. Supercedes ``order``, so
        if both are provided, fields in ``leading`` will appear first in the CSV file and
        remaining fields will appear in the order provided in ``order``.

    Returns
    -------
    None

    """
    if all([isinstance(s, Sequence) for s in sequences]):
        df = sequences_to_polars(
            sequences,
            annotations=annotations,
            columns=columns,
            properties=properties,
            drop_na_columns=drop_na_columns,
            order=order,
            exclude=exclude,
            leading=leading,
        )
    elif all([isinstance(s, Pair) for s in sequences]):
        df = pairs_to_polars(
            sequences,
            annotations=annotations,
            columns=columns,
            properties=properties,
            sequence_properties=sequence_properties,
            drop_na_columns=drop_na_columns,
            order=order,
            exclude=exclude,
            leading=leading,
        )
    else:
        raise ValueError(
            "All elements in the input list must be of the same type (either Sequence or Pair)."
        )

    df.write_csv(csv_file, separator=separator, include_header=header)


def read_airr(
    airr_file: str,
    match: Optional[dict] = None,
    fields: Optional[Iterable] = None,
    id_key: str = "sequence_id",
    sequence_key: str = "sequence",
    as_dataframe: bool = False,
) -> Iterable[Union[Sequence, Pair]]:
    """
    Reads a CSV file and returns a list of ``Sequence`` or ``Pair`` objects.

    Parameters
    ----------
    csv_file : str
        Path to the CSV file to be read.

    separator : str, default=","
        Column separator. Default is ``","``.

    match : dict, optional
        A ``dict`` for filtering sequences from the input file.
        Sequences must match all conditions to be returned. For example,
        the following ``dict`` will filter out all sequences for which
        the ``'v_gene:0'`` field is not ``'IGHV1-2'``:

        .. code-block:: python

            {'v_gene:0': 'IGHV1-2'}

    fields : list, optional
        A list of fields to be read from the input file. If not provided, all fields will be read.

    id_key : str, default="sequence_id"
        The name of the column that contains the sequence IDs. Default is ``"sequence_id"``.

    sequence_key : str, default="sequence"
        The name of the column that contains the sequence data. Default is ``"sequence"``.

    as_dataframe : bool, default=False
        If ``True``, the function will return a ``polars.DataFrame`` object.

        .. note::
            If ``True``, ``fields`` will be used to select columns from the input file, but
            ``match`` will be ignored.

    Returns
    -------
    Iterable[Union[Sequence, Pair]] or polars.DataFrame
        A list of ``Sequence`` or ``Pair`` objects or a ``polars.DataFrame`` object.

    """
    return read_csv(
        airr_file,
        separator="\t",
        match=match,
        fields=fields,
        id_key=id_key,
        sequence_key=sequence_key,
        as_dataframe=as_dataframe,
    )


def to_airr(
    sequences: Iterable[Union[Sequence, Pair]],
    airr_file: str,
    annotations: Optional[Iterable[str]] = None,
    columns: Optional[Iterable] = None,
    properties: Optional[Iterable[str]] = None,
    sequence_properties: Optional[Iterable[str]] = None,
    drop_na_columns: bool = True,
    order: Optional[Iterable[str]] = None,
    exclude: Optional[Union[str, Iterable[str]]] = None,
    leading: Optional[Union[str, Iterable[str]]] = None,
) -> None:
    """
    Saves a list of ``Pair`` objects to a CSV file.

    Parameters
    ----------
    sequences : Iterable[Sequence, Pair]
        List of ``Sequence`` or ``Pair`` objects to be saved to a CSV file. Required.

    airr_file : str
        Path to the output AIRR file. Required.

    annotations : list, default=None
        A list of annotation fields to be included in the CSV file. For ``Sequence``
        objects, this refers to fields in the ``annotations`` field. For ``Pair`` objects,
        this refers to fields in the heavy and light chain annotations.

    columns : list, default=None
        Used only for ``Pair`` objects. A list of fields to be retained in the output CSV
        file. Fields should be column names in the input file, such as ``"sequence:0"``,
        ``"sequence:1"``, ``"name"``, etc. This option is provided to allow differential
        selection of heavy and light chain fields. For cases in which the same fields will
        be selected for both chains, it is recommended to use ``annotations`` instead.

    properties : list, default=None
        A list of properties to be included in the CSV file. If not provided, everything
        in the ``annotations`` field of each heavy/light chain will be included.

    sequence_properties : list, default=None
        A list of sequence properties to be included. Differs from ``properties``, which
        refers to properties of the ``Pair`` object. These properties are those of the
        heavy/light ``Sequence`` objects. Ignored if the input is a list of ``Sequence`` objects.

    drop_na_columns : bool, default=True
        If ``True``, columns with all ``NaN`` values will be dropped from the CSV file.
        Default is ``True``.

    order : list, default=None
        A list of fields in the order they should appear in the CSV file.

    exclude : str or list, default=None
        Field or list of fields to be excluded from the CSV file.

    leading : str or list, default=None
        Field or list of fields to appear first in the CSV file. Supercedes ``order``, so
        if both are provided, fields in ``leading`` will appear first in the CSV file and
        remaining fields will appear in the order provided in ``order``.

    Returns
    -------
    None

    """
    to_csv(
        sequences,
        airr_file,
        separator="\t",
        header=True,
        annotations=annotations,
        columns=columns,
        properties=properties,
        sequence_properties=sequence_properties,
        drop_na_columns=drop_na_columns,
        order=order,
        exclude=exclude,
        leading=leading,
    )


def read_parquet(
    parquet_file: str,
    match: Optional[dict] = None,
    fields: Optional[Iterable] = None,
    id_key: str = "sequence_id",
    sequence_key: str = "sequence",
    as_dataframe: bool = False,
) -> Iterable[Union[Sequence, Pair]]:
    """
    Reads a Parquet file and returns a list of ``Sequence`` or ``Pair`` objects.

    Parameters
    ----------
    parquet_file : str
        Path to the Parquet file to be read.

    match : dict, optional
        A ``dict`` for filtering sequences from the input file.
        Sequences must match all conditions to be returned. For example,
        the following ``dict`` will filter out all sequences for which
        the ``'v_gene:0'`` field is not ``'IGHV1-2'``:

        .. code-block:: python

            {'v_gene:0': 'IGHV1-2'}

    fields : list, optional
        A list of fields to be read from the input file. If not provided, all fields will be read.

    id_key : str, default="sequence_id"
        The name of the column that contains the sequence IDs. Default is ``"sequence_id"``.

    sequence_key : str, default="sequence"
        The name of the column that contains the sequence data. Default is ``"sequence"``.

    as_dataframe : bool, default=False
        If ``True``, the function will return a ``polars.DataFrame`` object.

        .. note::
            If ``True``, ``fields`` will be used to select columns from the input file, but
            ``match`` will be ignored.

    Returns
    -------
    Iterable[Union[Sequence, Pair]] or polars.DataFrame
        A list of ``Sequence`` or ``Pair`` objects or a ``polars.DataFrame`` object.

    """
    df = pl.read_parquet(parquet_file)
    if as_dataframe:
        if fields is not None:
            df = df.select(fields)
        return df
    if any([c in df.columns for c in ["sequence:0", "sequence_id:0"]]):
        return pairs_from_polars(
            df, match=match, fields=fields, id_key=id_key, sequence_key=sequence_key
        )
    else:
        return sequences_from_polars(
            df, match=match, fields=fields, id_key=id_key, sequence_key=sequence_key
        )


def to_parquet(
    sequences: Iterable[Union[Sequence, Pair]],
    parquet_file: str,
    annotations: Optional[Iterable[str]] = None,
    columns: Optional[Iterable] = None,
    properties: Optional[Iterable[str]] = None,
    sequence_properties: Optional[Iterable[str]] = None,
    drop_na_columns: bool = True,
    order: Optional[Iterable[str]] = None,
    exclude: Optional[Union[str, Iterable[str]]] = None,
    leading: Optional[Union[str, Iterable[str]]] = None,
) -> None:
    """
    Saves a list of ``Pair`` objects to a Parquet file.

    Parameters
    ----------
    sequences : Iterable[Sequence, Pair]
        List of ``Sequence`` or ``Pair`` objects to be saved to a Parquet file. Required.

    parquet_file : str
        Path to the output Parquet file. Required.

    annotations : list, default=None
        A list of annotation fields to be included in the Parquet file. For ``Sequence``
        objects, this refers to fields in the ``annotations`` field. For ``Pair`` objects,
        this refers to fields in the heavy and light chain annotations.

    columns : list, default=None
        Used only for ``Pair`` objects. A list of fields to be retained in the output Polars
        DataFrame. Fields should be column names in the input file, such as ``"sequence:0"``,
        ``"sequence:1"``, ``"name"``, etc. This option is provided to allow differential
        selection of heavy and light chain fields. For cases in which the same fields will
        be selected for both chains, it is recommended to use ``annotations`` instead.

    properties : list, default=None
        A list of properties to be included in the Parquet file. If not provided, everything
        in the ``annotations`` field of each heavy/light chain will be included.

    sequence_properties : list, default=None
        A list of sequence properties to be included. Differs from ``properties``, which
        refers to properties of the ``Pair`` object. These properties are those of the
        heavy/light ``Sequence`` objects. Ignored if the input is a list of ``Sequence`` objects.

    drop_na_columns : bool, default=True
        If ``True``, columns with all ``NaN`` values will be dropped from the Parquet file.
        Default is ``True``.

    order : list, default=None
        A list of fields in the order they should appear in the Parquet file.

    exclude : str or list, default=None
        Field or list of fields to be excluded from the Parquet file.

    leading : str or list, default=None
        Field or list of fields to appear first in the Parquet file. Supercedes ``order``, so
        if both are provided, fields in ``leading`` will appear first in the Parquet file and
        remaining fields will appear in the order provided in ``order``.

    Returns
    -------
    None

    """
    if all([isinstance(s, Sequence) for s in sequences]):
        df = sequences_to_polars(
            sequences,
            annotations=annotations,
            columns=columns,
            properties=properties,
            drop_na_columns=drop_na_columns,
            order=order,
            exclude=exclude,
            leading=leading,
        )
    elif all([isinstance(s, Pair) for s in sequences]):
        df = pairs_to_polars(
            sequences,
            annotations=annotations,
            columns=columns,
            properties=properties,
            sequence_properties=sequence_properties,
            drop_na_columns=drop_na_columns,
            order=order,
            exclude=exclude,
            leading=leading,
        )
    else:
        raise ValueError(
            "All elements in the input list must be of the same type (either Sequence or Pair)."
        )

    df.write_parquet(parquet_file)


def read_sequences(
    file=None,
    format="airr",
    sep="\t",
    fields=None,
    match=None,
    id_key="sequence_id",
    sequence_key="sequence",
    db=None,
    collection=None,
    mongodb_kwargs=None,
):
    """
    Reads sequence data from a file and returns ``abutils.Seqeunce`` objects.

    Args:
    -----

    file (str): path to a file containing sequence data in any of the supported formats.

    format (str): format of the sequence file. Supported formats are: ``'airr'``, ``'tabular'``, ``'fasta'``,
        ``'fastq'``, ``'json'`` and ``'mongodb'``. Default is ``'airr'``.

    sep (str): character used to separate fields in ``'tabular'`` input files. This option is
        only used when ``format`` is ``'tabular'``. Default is ``'\t'``, which conforms with the
        default format for AIRR-compatible sequence annotation.

    id_key (str): name of the field containing the sequence ID. Default is ``'sequence_id'``.

    sequence_key (str): name of the field containing the sequence. Default is ``'sequence'``.

    db (str): mongodb database to query for sequence information. Required if ``format`` is ``'mongodb'``.

    collection (str): mongodb collection to query for sequence information. Required if ``format`` is ``'mongodb'``.

    mongodb_kwargs (dict): dictionary containing additional keyword arguments that will be passed to
        ``abutils.io.from_mongodb``.


    Returns:
    --------

    A list of ``abutils.Sequence`` objects.
    """
    format = format.lower()
    if format == "json":
        return read_json(
            file, id_key=id_key, sequence_key=sequence_key, fields=fields, match=match
        )
    elif format == "fasta":
        return read_fasta(file)
    elif format == "fastq":
        return read_fastq(file)
    elif format == "airr":
        return read_airr(file, fields=fields, match=match)
    elif format == "tabular":
        return read_csv(
            file,
            delimiter=sep,
            id_key=id_key,
            sequence_key=sequence_key,
            fields=fields,
            match=match,
        )
    elif format == "mongodb":
        if any([db is None, collection is None]):
            error = '<db> and <collection> are required arguments if the data type is "mongodb".'
            raise ValueError(error)
        return from_mongodb(db, collection, **mongodb_kwargs)
    else:
        error = f'Format type "{format}"" is not supported. '
        error += 'Supported file types are "airr", "fasta", "json", and "tabular".'
        raise ValueError(error)



# #!/usr/bin/env python
# # filename: ssh_tunnel.py


# #
# # Copyright (c) 2015 Bryan Briney
# # License: The MIT license (http://opensource.org/licenses/MIT)
# #
# # Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# # and associated documentation files (the "Software"), to deal in the Software without restriction,
# # including without limitation the rights to use, copy, modify, merge, publish, distribute,
# # sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# # furnished to do so, subject to the following conditions:
# #
# # The above copyright notice and this permission notice shall be included in all copies or
# # substantial portions of the Software.
# #
# # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# # BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# # NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# # DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# #


# # Borrowed heavily from Paramiko:
# # https://github.com/paramiko/paramiko/blob/master/demos/forward.py


# import getpass
# import os
# import socket
# import select

# try:
#     import SocketServer
# except ImportError:
#     import socketserver as SocketServer

# import argparse
# import sys

# import paramiko

# SSH_PORT = 22
# DEFAULT_PORT = 27017


# class ForwardServer(socketserver.ThreadingTCPServer):
#     daemon_threads = True
#     allow_reuse_address = True


# class Handler(socketserver.BaseRequestHandler):
#     def handle(self):
#         try:
#             chan = self.ssh_transport.open_channel(
#                 "direct-tcpip",
#                 (self.chain_host, self.chain_port),
#                 self.request.getpeername(),
#             )
#         except Exception as e:
#             print(
#                 "Incoming request to {}:{} failed: {}".format(
#                     self.chain_host, self.chain_port, repr(e)
#                 )
#             )
#             return
#         if chan is None:
#             print(
#                 "Incoming request to {}:{} was rejected by the SSH server.".format(
#                     self.chain_host, self.chain_port
#                 )
#             )
#             return

#         print(
#             "Connected!  Tunnel open {} -> {} -> {}".format(
#                 self.request.getpeername(),
#                 chan.getpeername(),
#                 (self.chain_host, self.chain_port),
#             )
#         )
#         while True:
#             r, w, x = select.select([self.request, chan], [], [])
#             if self.request in r:
#                 data = self.request.recv(1024)
#                 if len(data) == 0:
#                     break
#                 chan.send(data)
#             if chan in r:
#                 data = chan.recv(1024)
#                 if len(data) == 0:
#                     break
#                 self.request.send(data)
#         peername = self.request.getpeername()
#         chan.close()
#         self.request.close()
#         print("Tunnel closed from {}".format(peername,))


# def forward_tunnel(local_port, remote_host, remote_port, transport):
#     # this is convoluted, but necessary to configure things for the Handler
#     # object.  (SocketServer doesn't give Handlers any way to access the outer
#     # server.)
#     class SubHander(Handler):
#         chain_host = remote_host
#         chain_port = remote_port
#         ssh_transport = transport

#     ForwardServer(("", local_port), SubHander).serve_forever()


# def get_host_port(spec, default_port):
#     "parse 'hostname:22' into a host and port, with the port optional"
#     args = (spec.split(":", 1) + [default_port])[:2]
#     args[1] = int(args[1])
#     return args[0], args[1]


# HELP = """\
# Set up a forward tunnel across an SSH server, using paramiko. A local port
# (given with -p) is forwarded across an SSH session to an address:port from
# the SSH server. This is similar to the openssh -L option.
# """


# def parse_arguments():
#     parser = argparse.ArgumentParser(
#         usage="usage: ssh_tunnel [options] <ssh-server>[:<server-port>]",
#         version="ssh_tunnel",
#         description=HELP,
#     )
#     parser.add_argument(
#         "ssh_server", nargs=1, help="SSH server, as <ssh-server>[:<ssh-port>]"
#     )
#     parser.add_argument(
#         "-p",
#         "--local-port",
#         action="store",
#         type=int,
#         dest="port",
#         default=DEFAULT_PORT,
#         help="local port to forward (default: {})".format(DEFAULT_PORT),
#     )
#     parser.add_argument(
#         "-u",
#         "--user",
#         action="store",
#         type=str,
#         dest="user",
#         default=getpass.getuser(),
#         help="username for SSH authentication (default: {})".format(getpass.getuser()),
#     )
#     parser.add_argument(
#         "-K",
#         "--key",
#         action="store",
#         type=str,
#         dest="keyfile",
#         default=None,
#         help="private key file to use for SSH authentication",
#     )
#     parser.add_argument(
#         "--no-key",
#         action="store_false",
#         dest="look_for_keys",
#         default=True,
#         help="don't look for or use a private key file",
#     )
#     parser.add_argument(
#         "-P",
#         "--password",
#         dest="readpass",
#         default=False,
#         action="store_true",
#         help="Use a password for SSH.",
#     )
#     parser.add_argument(
#         "-r",
#         "--remote",
#         action="store",
#         required=True,
#         type=str,
#         dest="remote",
#         default=None,
#         metavar="host:port",
#         help="remote host and port to forward to",
#     )
#     args = parser.parse_args()

#     if args.remote is None:
#         parser.error("Remote address required (-r).")

#     server_host, server_port = get_host_port(args.ssh_server[0], SSH_PORT)
#     remote_host, remote_port = get_host_port(args.remote, args.port)
#     return args, (server_host, server_port), (remote_host, remote_port)


# def main():
#     args, server, remote = parse_arguments()
#     if args.readpass:
#         password = getpass.getpass("Enter SSH password: ")

#     client = paramiko.SSHClient()
#     client.load_system_host_keys()
#     client.set_missing_host_key_policy(paramiko.WarningPolicy())

#     print("Connecting to ssh host {}:{} ...".format(server[0], server[1]))
#     try:
#         client.connect(
#             server[0],
#             server[1],
#             username=args.user,
#             key_filename=args.keyfile,
#             look_for_keys=args.look_for_keys,
#             password=password,
#         )
#     except Exception as e:
#         print("*** Failed to connect to {}:{}: {}".format(server[0], server[1], e))
#         sys.exit(1)

#     print("Now forwarding port {} to {}:{} ...".format(args.port, remote[0], remote[1]))
#     try:
#         forward_tunnel(args.port, remote[0], remote[1], client.get_transport())
#     except KeyboardInterrupt:
#         print("C-c: Port forwarding stopped.")
#         sys.exit(0)


# if __name__ == "__main__":
#     main()



# __all__ = ['alignment', 'cluster', 'codons', 'color', 'convert',
#            'decorators', 'germlines', 'jobs', 'log', 'mongodb',
#            'phylogeny', 'pipeline', 'progbar', 's3', 'ssh_tunnel']



"""
Smoke tests for the plotting module.

These tests verify that plotting functions run without error and return
expected types. They do not verify the visual correctness of the plots.
"""

import matplotlib
matplotlib.use("Agg")  # Use non-interactive backend for testing

import numpy as np
import pandas as pd
import pytest

import matplotlib.pyplot as plt
import matplotlib.axes


# ==================
#      FIXTURES
# ==================


@pytest.fixture
def categorical_data():
    """DataFrame with categorical data for bar plots."""
    np.random.seed(42)
    return pd.DataFrame(
        {
            "category": np.random.choice(["A", "B", "C", "D"], 100),
            "value": np.random.randint(1, 100, 100),
            "group": np.random.choice(["X", "Y"], 100),
        }
    )


@pytest.fixture
def scatter_data():
    """DataFrame for scatter plots."""
    np.random.seed(42)
    return pd.DataFrame(
        {
            "x": np.random.rand(50),
            "y": np.random.rand(50),
            "category": np.random.choice(["A", "B", "C"], 50),
            "size": np.random.randint(10, 100, 50),
        }
    )


@pytest.fixture
def heatmap_data():
    """DataFrame for heatmap plots."""
    np.random.seed(42)
    return pd.DataFrame(
        np.random.rand(5, 5),
        index=[f"row{i}" for i in range(5)],
        columns=[f"col{i}" for i in range(5)],
    )


@pytest.fixture
def kde_data():
    """Data for KDE plots."""
    np.random.seed(42)
    return pd.DataFrame(
        {
            "values": np.concatenate([
                np.random.normal(0, 1, 100),
                np.random.normal(3, 0.5, 100),
            ]),
            "category": ["A"] * 100 + ["B"] * 100,
        }
    )


@pytest.fixture
def donut_data():
    """Data for donut plots."""
    return pd.DataFrame(
        {
            "category": ["A", "B", "C", "D"],
            "value": [30, 25, 25, 20],
        }
    )


# ========================
#    BAR PLOT TESTS
# ========================


def test_bar_basic_smoke(categorical_data):
    """Smoke test: bar() with x parameter runs without error."""
    from ..plots.bar import bar

    ax = bar(x="category", data=categorical_data)
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close("all")


def test_bar_with_hue_smoke(categorical_data):
    """Smoke test: bar() with hue parameter runs without error."""
    from ..plots.bar import bar

    ax = bar(x="category", hue="group", data=categorical_data)
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close("all")


def test_bar_horizontal_smoke(categorical_data):
    """Smoke test: bar() with horizontal orientation runs without error."""
    from ..plots.bar import bar

    ax = bar(y="category", data=categorical_data, orientation="horizontal")
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close("all")


# ==========================
#    SCATTER PLOT TESTS
# ==========================


def test_scatter_basic_smoke(scatter_data):
    """Smoke test: scatter() runs without error and returns Axes."""
    from ..plots.scatter import scatter

    ax = scatter(x="x", y="y", data=scatter_data)
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close("all")


def test_scatter_with_hue_smoke(scatter_data):
    """Smoke test: scatter() with hue parameter runs without error."""
    from ..plots.scatter import scatter

    ax = scatter(x="x", y="y", hue="category", data=scatter_data)
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close("all")


# ==========================
#    HEATMAP TESTS
# ==========================


@pytest.mark.xfail(reason="heatmap requires specific data format via PlotData")
def test_heatmap_basic_smoke(heatmap_data):
    """Smoke test: heatmap() runs without error and returns Axes."""
    from ..plots.heatmap import heatmap

    # heatmap expects a simple 2D array/DataFrame of values
    ax = heatmap(data=heatmap_data.values.tolist())
    assert ax is not None
    plt.close("all")


# =====================
#    KDE PLOT TESTS
# =====================


def test_kde_basic_smoke(kde_data):
    """Smoke test: kde() runs without error and returns Axes."""
    from ..plots.kde import kde

    ax = kde(x="values", data=kde_data)
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close("all")


def test_kde_with_hue_smoke(kde_data):
    """Smoke test: kde() with hue parameter runs without error."""
    from ..plots.kde import kde

    ax = kde(x="values", hue="category", data=kde_data)
    assert isinstance(ax, matplotlib.axes.Axes)
    plt.close("all")


# ======================
#    RIDGE PLOT TESTS
# ======================


def test_ridge_basic_smoke(kde_data):
    """Smoke test: ridge() runs without error."""
    from ..plots.ridge import ridge

    # ridge uses 'categories' and 'values' parameters
    ax = ridge(categories="category", values="values", data=kde_data, show=False)
    # ridge may return axes or figure depending on implementation
    assert ax is not None
    plt.close("all")


# ======================
#    DONUT PLOT TESTS
# ======================


def test_donut_basic_smoke(donut_data):
    """Smoke test: donut() runs without error."""
    from ..plots.donut import donut

    # donut uses 'values' for categories and 'counts' for values
    ax = donut(values="category", counts="value", data=donut_data)
    assert ax is not None
    plt.close("all")



#!/usr/bin/env python
# filename: similarity.py


#
# Copyright (c) 2023 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import itertools
import math
from typing import Iterable, Optional, Union

import numpy as np
import pandas as pd

from ..core.pair import Pair
from ..core.sequence import Sequence

__all__ = [
    "RepertoireSimilarity",
    "RepertoireSimilarities",
    "repertoire_similarity",
    # "normalize",
    # "make_continuous",
    "morisita_horn",
    "kullback_leibler",
    "jensen_shannon",
    "jaccard_similarity",
    "renkonen",
    "bray_curtis",
    "cosine_similarity",
]


class RepertoireSimilarity:
    """
    Object containing repertoire similarity data for a single pairwise comparison.
    """

    def __init__(
        self,
        similarities: Iterable[float],
        method: str,
        sample1_name: Optional[str] = None,
        sample2_name: Optional[str] = None,
    ):
        self.similarities = similarities
        self.method = method
        self.sample1_name = sample1_name if sample1_name is not None else "sample1"
        self.sample2_name = sample2_name if sample2_name is not None else "sample2"
        self._df = None

    def __repr__(self):
        return "<RepertoireSimilarity: {}>".format(self.method)

    def __str__(self):
        return "<RepertoireSimilarity: {}>".format(self.method)

    def __len__(self):
        return len(self.similarities)

    @property
    def df(self):
        """
        Returns a pandas ``DataFrame`` of similarity data.
        """
        if self._df is None:
            data = []
            for i, s in enumerate(self.similarities, 1):
                d = {
                    "sample1": self.sample1_name,
                    "sample2": self.sample2_name,
                    "method": self.method,
                    "iteration": i,
                    "similarity": s,
                }
                data.append(d)
            self._df = pd.DataFrame(data)
        return self._df

    @property
    def mean(self):
        """
        Returns the mean similarity value (if multiple iterations were performed)
        """
        return np.mean(self.similarities)

    @property
    def median(self):
        """
        Returns the median similarity value (if multiple iterations were performed)
        """
        return np.median(self.similarities)

    @property
    def min(self):
        """
        Returns the minimum similarity value (if multiple iterations were performed)
        """
        return np.min(self.similarities)

    @property
    def max(self):
        """
        Returns the maximum similarity value (if multiple iterations were performed)
        """
        return np.max(self.similarities)

    @property
    def std(self):
        """
        Returns the similarity standard deviation (if multiple iterations were performed)
        """
        return np.std(self.similarities)

    @property
    def sem(self):
        """
        Returns the similarity standard error of means (if multiple iterations were performed)
        """
        return np.std(self.similarities) / np.sqrt(len(self.similarities))


class RepertoireSimilarities:
    """
    Object containing repertoire similarity data for a multiple pairwise comparisons.
    """

    def __init__(self, similarities: Optional[Iterable[RepertoireSimilarity]] = None):
        self.similarities = similarities if similarities is not None else []
        self.df_needs_update = True
        self._df = None

    def __iter__(self):
        for s in self.similarities:
            yield s

    def __len__(self):
        return len(self.similarities)

    def __repr__(self):
        return "<RepertoireSimilarities: {} comparisons>".format(len(self))

    def __str__(self):
        return "<RepertoireSimilarities: {} comparisons>".format(len(self))

    def __add__(self, other):
        if isinstance(other, RepertoireSimilarity):
            self.add(other)
            return self
        elif isinstance(other, RepertoireSimilarities):
            for s in other:
                self.add(s)
            return self
        else:
            raise TypeError(
                "RepertoireSimilarities can only be added to RepertoireSimilarity or RepertoireSimilarities objects."
            )

    @property
    def df(self):
        """
        Returns a pandas ``DataFrame`` of similarity data.
        """
        if self._df is None or self.df_needs_update:
            if self.similarities:
                self._df = pd.concat([s.df for s in self.similarities])
                self.df_needs_update = False
        return self._df

    @property
    def means(self) -> np.array:
        """
        Returns an array of mean similarity values, one for each pairwise comparison.
        """
        return np.array([s.mean for s in self.similarities])

    @property
    def medians(self) -> np.array:
        """
        Returns an array of median similarity values, one for each pairwise comparison.
        """
        return np.array([s.median for s in self.similarities])

    @property
    def mins(self) -> np.array:
        """
        Returns an array of minimum similarity values, one for each pairwise comparison.
        """
        return np.array([s.min for s in self.similarities])

    @property
    def maxs(self) -> np.array:
        """
        Returns an array of maximum similarity values, one for each pairwise comparison.
        """
        return np.array([s.max for s in self.similarities])

    @property
    def stds(self) -> np.array:
        """
        Returns an array of standard deviations, one for each pairwise comparison.
        """
        return np.array([s.std for s in self.similarities])

    @property
    def sems(self) -> np.array:
        """
        Returns an array of standard errors of the mean, one for each pairwise comparison.
        """
        return np.array([s.sem for s in self.similarities])

    def squareform(
        self,
        # method=None,
        agg=np.mean,
    ) -> pd.DataFrame:
        """
        Returns a squareform representation of the similarity data.

        Parameters
        ----------
        method : str, optional
            The similarity method to use, if multiple similarity methods are present in the
            ``RepertoireSimilarities`` object. If ``None``, and multiple methods exist, all methods
            will be pooled (not recommended). Default is ``None``, which is the preferred
            argument when the data contains only a single similarity method.

        agg : function, optional
            The function used to aggregate similarity values. Default is ``np.mean``.
        """
        sq_data = {}
        for s1 in self.df["sample1"].unique():
            for s2 in self.df["sample2"].unique():
                _df = self.df[(self.df["sample1"] == s1) & (self.df["sample2"] == s2)]
                # if method is not None:
                #     _df = _df[_df["method"] == method]
                if _df.shape[0] == 0:
                    continue
                similarity = agg(_df["similarity"])
                if s1 not in sq_data:
                    sq_data[s1] = {}
                sq_data[s1][s2] = similarity
        return pd.DataFrame(sq_data)

    def add(self, similarity):
        """
        Adds a RepertoireSimilarity object.

        Parameters
        ----------
        similarity : RepertoireSimilarity
            A RepertoireSimilarity object.

        """
        self.similarities.append(similarity)
        self.df_needs_update = True


def repertoire_similarity(
    repertoires: Iterable[Union[Iterable[Sequence], Iterable[Pair]]],
    names: Optional[Iterable[str]] = None,
    method: str = "morisita-horn",
    features: Union[str, Iterable[str], None] = None,
    n_iters: int = 1,
    subsample_size: Optional[int] = None,
    sample_with_replacement: bool = False,
    pairs_only: bool = False,
    chain: Optional[str] = None,
    force_self_comparisons: bool = False,
    seed: Union[int, float, str, None] = None,
) -> Union[float, RepertoireSimilarity, RepertoireSimilarities]:
    """
    Compute the pairwise similarity between two or more repertoires.

    Parameters
    ----------
    repertoires : list
        A list of repertoires. Each repertoire can be a list of ``abutils.Sequence``
        objects or a list of ``abutils.Pair`` objects.

        .. note::
            Individual repertoires cannot mix ``Sequence`` and ``Pair`` objects,
            but the list of repertoires can contain both. For example, comparing a
            repetoire of ``Sequence`` objects to a repertoire of ``Pair`` objects is
            allowed.

    names : list, optional
        A list of names for the repertoires. If ``None``, names will be generated as
        ``"repertoire1"``, ``"repertoire2"``, etc.

    method : str, optional
        The similarity method to use. Options are:
            - ``"morisita-horn"``
            - ``"kullback-leibler"``
            - ``"jensen-shannon"``
            - ``"jaccard"``
            - ``"bray-curtis"``
            - ``"renkonen"``
            - ``"cosine"``
        Default is ``"morisita-horn"``.

    features : str, list, optional
        The features to use for similarity calculation. Default is ``["v_gene", "j_gene", "cdr3_length"]``.

    n_iters : int, optional
        The number of iterations to perform. Default is ``1``.

    subsample_size : int, optional
        The number of sequences to subsample from each repertoire. If ``None``, the smallest repertoire
        will be used as the subsample size. Default is ``None``.

    sample_with_replacement : bool, optional
        Whether to subsample with replacement. Default is ``False``.

    pairs_only : bool, optional
        Whether to use only paired sequences. Requires that all repertoires contain ``Pair`` objects,
        rather than ``Sequence`` objects. Default is ``False``.

    chain : str, optional
        The chain to use for similarity calculation. Default is ``None``, which will use all chains.

        .. note::
            This applies to both ``Sequence`` and ``Pair`` objects. If a repertoire contains
            ``Sequence`` objects from multiple chains, only sequences of the specified chain will
            be used for similarity calculation, and the rest will be ignored.

    force_self_comparisons : bool, optional
        Whether to force self-comparisons. Only used if exactly two repertoires are provided.
        Default is ``False``, which performs only a single pairwise comparison when exactly
        two repertoires are provided. If more than two repertoires are provided, all pairwise
        comparisons will be performed, including self-comparisons.

    seed : int, float, str, optional
        The seed to use for random number generation. Default is ``None``.

    Returns
    -------
    float or RepertoireSimilarity or RepertoireSimilarities
        If only two repertoires are provided and `n_iters` is ``1``, the similarity value will be returned as a float.
        If only two repertoires are provided and `n_iters` is greater than ``1``, a ``RepertoireSimilarity`` object will be
        returned. If more than two repertoires are provided, a ``RepertoireSimilarities`` object will be
        returned.

    """
    # features
    if features is None:
        features = ["v_gene", "j_gene", "cdr3_length"]
    elif isinstance(features, str):
        features = [features]
    # chain/locus
    chain_dict = {
        "heavy": ["IGH"],
        "kappa": ["IGK"],
        "lambda": ["IGL"],
        "light": ["IGK", "IGL"],
        "all": ["IGH", "IGK", "IGL"],
    }
    if chain is None:
        locus = chain_dict["all"]
    elif isinstance(chain, str):
        locus = chain_dict.get(chain.lower(), [chain])
    else:
        locus = chain
    # subsample size
    if subsample_size is None:
        subsample_size = min([len(r) for r in repertoires])
    # names
    if names is None:
        names = [f"repertoire{i}" for i, _ in enumerate(repertoires, 1)]
    elif len(names) != len(repertoires):
        err = "\nThe number of names must match the number of repertoires.\n"
        err += f"You provided {len(names)} names and {len(repertoires)} repertoires.\n"
        raise RuntimeError(err)
    # process repertoires
    if (n := len(repertoires)) < 2:
        err = f"\nYou must provide at least two repertoires for comparison. You provided {n}\n"
        raise RuntimeError(err)
    rep_dict = {n: r for n, r in zip(names, repertoires)}
    # similarities
    similarities = []
    if len(names) == 2 and not force_self_comparisons:
        comps = [names]
    else:
        comps = itertools.product(names, repeat=2)
    for name1, name2 in comps:
        repertoire1 = rep_dict[name1]
        repertoire2 = rep_dict[name2]
        sim = _calculate_similarity(
            repertoire1=repertoire1,
            repertoire2=repertoire2,
            name1=name1,
            name2=name2,
            method=method,
            features=features,
            subsample_size=subsample_size,
            with_replacement=sample_with_replacement,
            n_iters=n_iters,
            pairs_only=pairs_only,
            locus=locus,
            seed=seed,
        )
        similarities.append(sim)
    # single comparison, single iteration
    if len(similarities) == 1 and n_iters == 1:
        return similarities[0].similarities[0]
    # single comparison, multiple iterations
    elif len(similarities) == 1:
        return similarities[0]
    # multiple comparisons
    else:
        return RepertoireSimilarities(similarities)


def _calculate_similarity(
    repertoire1: Union[Iterable[Sequence], Iterable[Pair]],
    repertoire2: Union[Iterable[Sequence], Iterable[Pair]],
    name1: str,
    name2: str,
    method: str,
    features: Iterable[str],
    subsample_size: int,
    with_replacement: bool,
    n_iters: int,
    pairs_only: bool,
    locus: Iterable[str],
    seed: Union[int, float, str],
) -> Union[float, RepertoireSimilarity]:
    """
    Calculates the similarity between two repertoires.

    Parameters
    ----------
    repertoire1 : list
        A list of ``abutils.Sequence`` or ``abutils.Pair`` objects.

    repertoire2 : list
        A list of ``abutils.Sequence`` or ``abutils.Pair`` objects.

    name1 : str
        The name of repertoire1.

    name2 : str
        The name of repertoire2.

    method : str
        The similarity method to use.

    features : list
        The features to use for similarity calculation.

    subsample_size : int
        The number of sequences to subsample from each repertoire.

    with_replacement : bool
        Whether to subsample with replacement.

    n_iters : int
        The number of iterations to perform.

    pairs_only : bool
        Whether to use only paired sequences.

    locus : list
        The locus to use for similarity calculation.

    seed : int, float, str
        The seed to use for random number generation.

    Returns
    -------
    RepertoireSimilarity
        A RepertoireSimilarity object.

    """
    # get features
    features1 = _get_features(
        repertoire=repertoire1, features=features, locus=locus, pairs_only=pairs_only
    )
    features2 = _get_features(
        repertoire=repertoire2, features=features, locus=locus, pairs_only=pairs_only
    )
    # seed
    if seed is not None:
        np.random.seed(seed)
    # get similarity function
    similarity_funcs = {
        "morisita-horn": morisita_horn,
        "kullback-leibler": kullback_leibler,
        "jensen-shannon": jensen_shannon,
        "jaccard": jaccard_similarity,
        "bray-curtis": bray_curtis,
        "renkonen": renkonen,
        "cosine": cosine_similarity,
    }
    similarity_func = similarity_funcs.get(method.lower(), None)
    if similarity_func is None:
        err = f"\nThe provided similarity method ({method}) is not supported.\n"
        err += "Supported methods are:\n - "
        err += "\n - ".join(similarity_funcs.keys())
        raise RuntimeError(err)
    # calculate similarity
    similarities = []
    for i in range(n_iters):
        # subsample
        subsample1 = np.random.choice(
            features1, size=subsample_size, replace=with_replacement
        )
        subsample2 = np.random.choice(
            features2, size=subsample_size, replace=with_replacement
        )
        # count features
        counts1 = pd.Series(subsample1).value_counts()
        counts2 = pd.Series(subsample2).value_counts()
        df = pd.DataFrame({"sample1": counts1, "sample2": counts2}).fillna(0)
        # similarity
        similarity = similarity_func(df["sample1"], df["sample2"])
        similarities.append(similarity)
    return RepertoireSimilarity(similarities, method, name1, name2)


def _get_features(
    repertoire: Union[Iterable[Sequence], Iterable[Pair]],
    features: Iterable[str],
    locus: Iterable[str],
    pairs_only: bool = False,
) -> Iterable[str]:
    """
    Gets the requested features from a repertoire.

    Parameters
    ----------
    repertoire : list
        A list of ``abutils.Sequence`` or ``abutils.Pair`` objects.

    features : list
        The features to use for similarity calculation.

    locus : list
        The locus to use for similarity calculation.

    pairs_only : bool
        Whether to use only paired sequences.

    Returns
    -------
    list
        A list of features.

    """
    all_features = []
    # process Pair objects
    if all([isinstance(r, Pair) for r in repertoire]):
        if pairs_only:
            repertoire = [r for r in repertoire if r.is_pair]
        for r in repertoire:
            feat_list = []
            if "IGH" in locus and r.heavy is not None:
                for f in features:
                    feat = r.heavy[f]
                    feat_list.append(str(feat) if feat is not None else "NA")
            if r.light is not None and r.light["locus"] in locus:
                for f in features:
                    feat = r.light[f]
                    feat_list.append(str(feat) if feat is not None else "NA")
            all_features.append("__".join(feat_list))
    else:
        for r in repertoire:
            feat_list = []
            if r["locus"] in locus:
                for f in features:
                    feat = r[f]
                    feat_list.append(str(feat) if feat is not None else "NA")
            all_features.append("__".join(feat_list))
    return all_features


def normalize(
    s1: Iterable[Union[int, float]], s2: Iterable[Union[int, float]]
) -> tuple:
    """
    Normalizes two distributions.

    Parameters
    ----------
    s1 : list
        An iterable of feature frequency values from sample 1.

    s2 : list
        An iterable of feature frequency values from sample 2.

    Returns
    -------
    tuple
        A tuple of normalized distributions.

    """
    return s1 / np.sum(s1), s2 / np.sum(s2)


def make_continuous(
    s1: Iterable[Union[int, float]], s2: Iterable[Union[int, float]]
) -> tuple:
    """
    Makes two distributions continuous by removing zero values.

    Parameters
    ----------
    s1 : list
        An iterable of feature frequency values from sample 1.

    s2 : list
        An iterable of feature frequency values from sample 2.

    Returns
    -------
    tuple
        A tuple of continuous distributions.

    """
    df = pd.DataFrame({"s1": s1, "s2": s2})
    df = df.fillna(0)
    df = df[(df["s1"] != 0) & (df["s2"] != 0)]
    return df["s1"], df["s2"]


# -----------------------------------
#       SIMILARITY METHODS
# -----------------------------------


def morisita_horn(sample1: Iterable, sample2: Iterable) -> float:
    """
    Calculates the Marista-Horn similarity for two distributions.

    Parameters
    ----------
    sample1 : list
        An iterable of feature frequency values from sample 1.

    sample2 : list
        An iterable of feature frequency values from sample 2.

    Returns
    -------
    float
        The Morisita-Horn similarity between sample1 and sample2.

    """
    X = sum(sample1)
    Y = sum(sample2)
    XY = X * Y
    sumXiYi = 0
    sumXiSq = 0
    sumYiSq = 0
    for x, y in zip(sample1, sample2):
        sumXiYi += x * y
        sumXiSq += x * x
        sumYiSq += y * y
    numerator = 2 * sumXiYi
    denominator = (float(sumXiSq) / (X * X) + float(sumYiSq) / (Y * Y)) * XY
    return numerator / denominator


def kullback_leibler(sample1: Iterable, sample2: Iterable) -> float:
    """
    Calculates the Kullback-Leibler similarity for two distributions.

    .. note::
        This is not a true metric, but is included for completeness.
        Kullback-Leibler similarity is not symmetric, so the order of the samples
        matters. KL(A, B) != KL(B, A).

    Parameters
    ----------
    sample1 : list
        An iterable of feature frequency values from sample 1.

    sample2 : list
        An iterable of feature frequency values from sample 2.

    Returns
    -------
    float
        The Morisita-Horn similarity between sample1 and sample2.

    """
    cont1, cont2 = make_continuous(sample1, sample2)
    norm1, norm2 = normalize(cont1, cont2)
    kl = 0
    for Pi, Qi in zip(norm1, norm2):
        kl += Pi * math.log(Pi / Qi)
    return 1 - kl


def jensen_shannon(sample1: Iterable, sample2: Iterable) -> float:
    """
    Calculates the Jensen-Shannon similarity for two distributions.

    Parameters
    ----------
    sample1 : list
        An iterable of feature frequency values from sample 1.

    sample2 : list
        An iterable of feature frequency values from sample 2.

    Returns
    -------
    float
        The Jensen-Shannon similarity between sample1 and sample2.

    """
    norm1, norm2 = normalize(sample1, sample2)
    M = [(n1 + n2) / 2 for n1, n2 in zip(norm1, norm2)]
    js = 0.5 * (kullback_leibler(norm1, M) + kullback_leibler(norm2, M))
    return 1 - js


def jaccard_similarity(sample1: Iterable, sample2: Iterable) -> float:
    """
    Calculates the Jaccard similarity for two distributions.

    Parameters
    ----------
    sample1 : list
        An iterable of feature frequency values from sample 1.

    sample2 : list
        An iterable of feature frequency values from sample 2.

    Returns
    -------
    float
        The Jaccard similarity between sample1 and sample2.

    """
    intersection = 0
    union = 0
    for s1, s2 in zip(sample1, sample2):
        intersection += min(s1, s2)
        union += max(s1, s2)
    return intersection / union


def renkonen(sample1: Iterable, sample2: Iterable) -> float:
    """
    Calculates the Renkonen similarity for two distributions.

    Parameters
    ----------
    sample1 : list
        An iterable of feature frequency values from sample 1.

    sample2 : list
        An iterable of feature frequency values from sample 2.

    Returns
    -------
    float
        The Renkonen similarity between sample1 and sample2.

    """
    norm1, norm2 = normalize(sample1, sample2)
    renkonen = sum([min(vals) for vals in zip(norm1, norm2)])
    return renkonen


def bray_curtis(sample1: Iterable, sample2: Iterable) -> float:
    """
    Calculates the Bray-Curtis similarity for two distributions.

    Parameters
    ----------
    sample1 : list
        An iterable of feature frequency values from sample 1.

    sample2 : list
        An iterable of feature frequency values from sample 2.

    Returns
    -------
    float
        The Bray-Curtis similarity between sample1 and sample2.

    """
    norm1, norm2 = normalize(sample1, sample2)
    sum_diff = sum(abs(x - y) for x, y in zip(norm1, norm2))
    sum_sum = sum(x + y for x, y in zip(norm1, norm2))
    bray_curtis = sum_diff / sum_sum
    return bray_curtis


def cosine_similarity(sample1: Iterable, sample2: Iterable) -> float:
    """
    Calculates the cosine similarity for two distributions.

    Parameters
    ----------
    sample1 : list
        An iterable of feature frequency values from sample 1.

    sample2 : list
        An iterable of feature frequency values from sample 2.

    Returns
    -------
    float
        The cosine similarity between sample1 and sample2.

    """
    numerator = sum(x * y for x, y in zip(sample1, sample2))
    denominator = math.sqrt(sum(x * x for x in sample1)) * math.sqrt(
        sum(y * y for y in sample2)
    )
    cosine = numerator / denominator
    return cosine



#!/usr/bin/env python
# filename: decorators.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals


def lazy_property(func):
    """
    Wraps a property to provide lazy evaluation. Eliminates boilerplate.
    Also provides for setting and deleting the property.

    Use as you would use the @property decorator::

        # OLD:
        class MyClass():
            def __init__():
                self._compute = None

            @property
            def compute(self):
                if self._compute is None:
                    # computationally intense stuff
                    # ...
                    # ...
                    self._compute = result
                return self._compute

            @compute.setter
            def compute(self, value):
                self._compute = value


        # NEW:
        class MyClass():

            def __init__():
                pass

            @lazy_property
            def compute(self):
                # computationally intense stuff
                # ...
                # ...
                return result

    .. note:

        Properties wrapped with ``lazy_property`` are only evaluated once.
        If the instance state changes, lazy properties will not be automatically
        re-evaulated and the update must be explicitly called for::

            c = MyClass(data)
            prop = c.lazy_property

            # If you update some data that affects c.lazy_property
            c.data = new_data

            # c.lazy_property won't change
            prop == c.lazy_property  # TRUE

            # If you want to update c.lazy_property, you can delete it, which will
            # force it to be recomputed (with the new data) the next time you use it
            del c.lazy_property
            new_prop = c.lazy_property
            new_prop == prop  # FALSE
    """
    attr_name = "_lazy_" + func.__name__

    @property
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, func(self))
        return getattr(self, attr_name)

    @_lazy_property.deleter
    def _lazy_property(self):
        if hasattr(self, attr_name):
            delattr(self, attr_name)

    @_lazy_property.setter
    def _lazy_property(self, value):
        setattr(self, attr_name, value)

    return _lazy_property


def coroutine(func):
    """
    Initializes a coroutine -- essentially it just takes a
    generator function and calls generator.next() to get
    things going.
    """

    def start(*args, **kwargs):
        cr = func(*args, **kwargs)
        cr.next()
        return cr

    return start
